!
!  Copyright (C) 2017 College of Engineering, Swansea University
!
!  This file is part of the SwanSim FLITE suite of tools.
!
!  SwanSim FLITE is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  SwanSim FLITE is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this SwanSim FLITE product. 
!  If not, see <http://www.gnu.org/licenses/>.
!



MODULE HighOrderCell

  USE JACOBIANS

  TYPE :: HighOrderCellType
     INTEGER :: Nsd           = 0
     INTEGER :: GridOrder     = 1
     INTEGER :: numCellNodes  = 0
     INTEGER :: numFaceNodes  = 0
     INTEGER :: numVertices   = 0
     INTEGER :: optionNodes   = 1
     INTEGER :: numGaussNodes = 0
     REAL*8,  DIMENSION(:,:),   POINTER :: Coord => null()
     INTEGER, DIMENSION(:,:),   POINTER :: faceNodes => null()
     REAL*8,  DIMENSION(:,:),   POINTER :: gaussShapeFunct => null()
     REAL*8,  DIMENSION(:,:,:), POINTER :: gaussShapeDeriv => null()
     REAL*8,  DIMENSION(:),     POINTER :: gaussWeights => null()
     REAL*8,  DIMENSION(:,:),   POINTER :: shapeMax => null()
     INTEGER, DIMENSION(:),     POINTER :: shapeLU => null()
  END TYPE HighOrderCellType


CONTAINS

  FUNCTION HighOrderCell_numCellNodes(p, Nsd) RESULT(N)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, Nsd
    INTEGER :: N
    IF(Nsd==1)THEN
       N = p+1
    ELSE IF(Nsd==2)THEN
       N = (p+1) * (p+2) / 2
    ELSE IF(Nsd==3)THEN
       N = (p+1) * (p+2) * (p+3) / 6
    ELSE
       STOP 'cmewpfjk92-4jg-g'
    ENDIF
  END FUNCTION HighOrderCell_numCellNodes


  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ! Modified routine
  ! lglnodes.m
  !
  ! Computes the Legendre-Gauss-Lobatto nodes, weights and the LGL Vandermonde 
  ! matrix. The LGL nodes are the zeros of (1-x^2)*P'_N(x). Useful for numerical
  ! integration and spectral methods. 
  !
  ! Reference on LGL nodes and weights: 
  !   C. Canuto, M. Y. Hussaini, A. Quarteroni, T. A. Tang, "Spectral Methods
  !   in Fluid Dynamics," Section 2.3. Springer-Verlag 1987
  !
  ! Written by Greg von Winckel - 04/17/2004
  ! Contact: gregvw@chtm.unm.edu
  !
  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  SUBROUTINE HighOrderCell_feketeNodes1D(p, x)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    REAL*8, DIMENSION(*), INTENT(out) :: x
    REAL*8, PARAMETER :: PI = 3.141592653589793D0
    REAL*8, PARAMETER :: eps = 2.2204D-16
    REAL*8 :: c
    REAL*8, DIMENSION(:,:), POINTER :: bP
    REAL*8, DIMENSION(:  ), POINTER :: xold

    INTEGER :: i, N1, k

    ! Truncation + 1
    N1=p+1

    ! Use the Chebyshev-Gauss-Lobatto nodes as the first guess
    DO i = 1,N1
       c = PI - (i-1.d0)/p * PI
       x(i) = dcos(c)
    ENDDO

    ! The Legendre Vandermonde Matrix
    ALLOCATE(bP(N1,N1), xold(N1))
    bP = 0

    ! Compute bP_(p) using the recursion relation
    ! Compute its first and second derivatives and 
    ! update x using the Newton-Raphson method.
    xold=2

    DO WHILE (MAXVAL(ABS(x(1:N1)-xold(1:N1)))>eps)
       xold(1:N1) = x(1:N1)

       bP(1:N1,1)=1    
       bP(1:N1,2)=x(1:N1)
       DO k=2,p
          DO i = 1, N1
             bP(i,k+1)=( (2*k-1.d0)* x(i)*bP(i,k) - (k-1.d0)*bP(i,k-1) ) /k
          ENDDO
       ENDDO

       DO i = 1,N1
          x(i)=xold(i)-( x(i)*bP(i,N1) -bP(i,p) ) / ( N1*bP(i,N1) )
       ENDDO
    ENDDO
    
    !--- make the first two nodes are the two ends
    xold(1:N1-2) = x(2:N1-1)
    x(2)         = x(N1)
    x(3:N1)      = xold(1:N1-2)

    DEALLOCATE(bP, xold)
  END SUBROUTINE HighOrderCell_feketeNodes1D


  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  SUBROUTINE HighOrderCell_feketeNodes2D(p, p2D)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    REAL*8,  INTENT(out) :: p2D(2,*)
    INTEGER :: ND

    ND = (p+1)*(p+2)/2

    SELECT CASE(p)
    CASE (1)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000,  1.000000000000D+000    &
            /) , (/2,ND/) )

    CASE (2)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000,  1.000000000000D+000,   &
            0.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, -1.354566102931D-018,   &
            2.775557561563D-017,  0.000000000000D+000    &
            /) , (/2,ND/) )

    CASE (3)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000,  1.000000000000D+000,   &
            -4.472135955000D-001, -1.000000000000D+000,   &
            4.472135955000D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -4.472135955000D-001,   &
            -3.333333333333D-001, -3.333333333333D-001,   &
            4.472135949353D-001, -4.472135949353D-001,   &
            -1.000000000000D+000,  4.472135955000D-001,   &
            -4.472135949353D-001,  4.472135949353D-001    &
            /) , (/2,ND/) )

    CASE (4)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000,  1.000000000000D+000,   &
            -6.546536707080D-001, -1.000000000000D+000,   &
            1.096484628801D-016, -1.000000000000D+000,   &
            6.546536707080D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -6.546536707080D-001,   &
            -5.669152707560D-001, -5.669152707560D-001,   &
            1.338305407411D-001, -5.669152702923D-001,   &
            6.546536707063D-001, -6.546536707063D-001,   &
            -1.000000000000D+000,  3.185439549414D-017,   &
            -5.669152702923D-001,  1.338305407411D-001,   &
            -1.289518317015D-016,  1.999282576988D-016,   &
            -1.000000000000D+000,  6.546536707080D-001,   &
            -6.546536707063D-001,  6.546536707063D-001    &
            /) , (/2,ND/) )

    CASE (5)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -7.650553239295D-001, -1.000000000000D+000,   &
            -2.852315164806D-001, -1.000000000000D+000,   &
            2.852315164806D-001, -1.000000000000D+000,   &
            7.650553239295D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -7.650553239295D-001,   &
            -7.039610574860D-001, -7.039610574860D-001,   &
            -1.583489221960D-001, -6.833021568502D-001,   &
            4.079221144413D-001, -7.039610573420D-001,   &
            7.650553238262D-001, -7.650553238262D-001,   &
            -1.000000000000D+000, -2.852315164806D-001,   &
            -6.833021568502D-001, -1.583489221960D-001,   &
            -1.583489213190D-001, -1.583489213190D-001,   &
            2.852315161321D-001, -2.852315161321D-001,   &
            -1.000000000000D+000, 2.852315164806D-001,   &
            -7.039610573420D-001, 4.079221144413D-001,   &
            -2.852315161321D-001, 2.852315161321D-001,   &
            -1.000000000000D+000, 7.650553239295D-001,   &
            -7.650553238262D-001, 7.650553238262D-001    &
            /) , (/2,ND/) )

    CASE (6)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -8.302238962786D-001, -1.000000000000D+000,   &
            -4.688487934707D-001, -1.000000000000D+000,   &
            1.045388153545D-016, -1.000000000000D+000,   &
            4.688487934707D-001, -1.000000000000D+000,   &
            8.302238962786D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -8.302238962786D-001,   &
            -7.873290633734D-001, -7.873290633734D-001,   &
            -3.674604088246D-001, -7.656381655817D-001,   &
            1.330985729989D-001, -7.656381655652D-001,   &
            5.746581261104D-001, -7.873290631189D-001,   &
            8.302238962418D-001, -8.302238962418D-001,   &
            -1.000000000000D+000, -4.688487934707D-001,   &
            -7.656381655817D-001, -3.674604088246D-001,   &
            -3.333333333019D-001, -3.333333333019D-001,   &
            1.330985734563D-001, -3.674604077452D-001,   &
            4.688487933245D-001, -4.688487933245D-001,   &
            -1.000000000000D+000, 3.151729351098D-017,   &
            -7.656381655652D-001, 1.330985729989D-001,   &
            -3.674604077452D-001, 1.330985734563D-001,   &
            -3.931027840594D-016, 3.983868829685D-016,   &
            -1.000000000000D+000, 4.688487934707D-001,   &
            -7.873290631189D-001, 5.746581261104D-001,   &
            -4.688487933245D-001, 4.688487933245D-001,   &
            -1.000000000000D+000, 8.302238962786D-001,   &
            -8.302238962418D-001, 8.302238962418D-001    &
            /) , (/2,ND/) )

    CASE (7)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -8.717401485096D-001, -1.000000000000D+000,   &
            -5.917001814331D-001, -1.000000000000D+000,   &
            -2.092992179025D-001, -1.000000000000D+000,   &
            2.092992179025D-001, -1.000000000000D+000,   &
            5.917001814331D-001, -1.000000000000D+000,   &
            8.717401485096D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -8.717401485096D-001,   &
            -8.408229863571D-001, -8.408229863571D-001,   &
            -5.132707143760D-001, -8.221139049920D-001,   &
            -9.202931298384D-002, -8.159413760606D-001,   &
            3.353846180094D-001, -8.221139049553D-001,   &
            6.816459722422D-001, -8.408229862340D-001,   &
            8.717401484820D-001, -8.717401484820D-001,   &
            -1.000000000000D+000, -5.917001814331D-001,   &
            -8.221139049920D-001, -5.132707143760D-001,   &
            -4.735817900919D-001, -4.735817900919D-001,   &
            -5.283642013648D-002, -4.735817898113D-001,   &
            3.353846187226D-001, -5.132707136842D-001,   &
            5.917001813187D-001, -5.917001813187D-001,   &
            -1.000000000000D+000, -2.092992179025D-001,   &
            -8.159413760606D-001, -9.202931298384D-002,   &
            -4.735817898113D-001, -5.283642013648D-002,   &
            -9.202931182453D-002, -9.202931182453D-002,   &
            2.092992177702D-001, -2.092992177702D-001,   &
            -1.000000000000D+000, 2.092992179025D-001,   &
            -8.221139049553D-001, 3.353846180094D-001,   &
            -5.132707136842D-001, 3.353846187226D-001,   &
            -2.092992177702D-001, 2.092992177702D-001,   &
            -1.000000000000D+000, 5.917001814331D-001,   &
            -8.408229862340D-001, 6.816459722422D-001,   &
            -5.917001813187D-001, 5.917001813187D-001,   &
            -1.000000000000D+000, 8.717401485096D-001,   &
            -8.717401484820D-001, 8.717401484820D-001    &
            /) , (/2,ND/) )

    CASE (8)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -8.997579954115D-001, -1.000000000000D+000,   &
            -6.771862795107D-001, -1.000000000000D+000,   &
            -3.631174638262D-001, -1.000000000000D+000,   &
            1.178993670744D-016, -1.000000000000D+000,   &
            3.631174638262D-001, -1.000000000000D+000,   &
            6.771862795107D-001, -1.000000000000D+000,   &
            8.997579954115D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -8.997579954115D-001,   &
            -8.768515641853D-001, -8.768515641853D-001,   &
            -6.164513258919D-001, -8.614874875968D-001,   &
            -2.655194740903D-001, -8.539081855183D-001,   &
            1.194276574448D-001, -8.539081855180D-001,   &
            4.779388123284D-001, -8.614874875985D-001,   &
            7.537031279712D-001, -8.768515640844D-001,   &
            8.997579953940D-001, -8.997579953940D-001,   &
            -1.000000000000D+000, -6.771862795107D-001,   &
            -8.614874875968D-001, -6.164513258919D-001,   &
            -5.795567713367D-001, -5.795567713367D-001,   &
            -2.163656669535D-001, -5.672686665101D-001,   &
            1.591135422664D-001, -5.795567711234D-001,   &
            4.779388129119D-001, -6.164513253284D-001,   &
            6.771862794386D-001, -6.771862794386D-001,   &
            -1.000000000000D+000, -3.631174638262D-001,   &
            -8.539081855183D-001, -2.655194740903D-001,   &
            -5.672686665101D-001, -2.163656669535D-001,   &
            -2.163656665564D-001, -2.163656665564D-001,   &
            1.194276586071D-001, -2.655194728995D-001,   &
            3.631174636994D-001, -3.631174636994D-001,   &
            -1.000000000000D+000, -4.844440267744D-017,   &
            -8.539081855180D-001, 1.194276574448D-001,   &
            -5.795567711234D-001, 1.591135422664D-001,   &
            -2.655194728995D-001, 1.194276586071D-001,   &
            -4.319307184444D-016, 3.221461381835D-016,   &
            -1.000000000000D+000, 3.631174638262D-001,   &
            -8.614874875985D-001, 4.779388123284D-001,   &
            -6.164513253284D-001, 4.779388129119D-001,   &
            -3.631174636994D-001, 3.631174636994D-001,   &
            -1.000000000000D+000, 6.771862795107D-001,   &
            -8.768515640844D-001, 7.537031279712D-001,   &
            -6.771862794386D-001, 6.771862794386D-001,   &
            -1.000000000000D+000, 8.997579954115D-001,   &
            -8.997579953940D-001, 8.997579953940D-001    &
            /) , (/2,ND/) )

    CASE (9)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.195339081665D-001, -1.000000000000D+000,   &
            -7.387738651055D-001, -1.000000000000D+000,   &
            -4.779249498104D-001, -1.000000000000D+000,   &
            -1.652789576664D-001, -1.000000000000D+000,   &
            1.652789576664D-001, -1.000000000000D+000,   &
            4.779249498104D-001, -1.000000000000D+000,   &
            7.387738651055D-001, -1.000000000000D+000,   &
            9.195339081665D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.195339081665D-001,   &
            -9.021308609852D-001, -9.021308609852D-001,   &
            -6.912196116129D-001, -8.896483841425D-001,   &
            -3.979515790090D-001, -8.822870241066D-001,   &
            -6.008247252297D-002, -8.798350575344D-001,   &
            2.802386010868D-001, -8.822870241144D-001,   &
            5.808679947612D-001, -8.896483841442D-001,   &
            8.042617216313D-001, -9.021308609041D-001,   &
            9.195339081537D-001, -9.195339081537D-001,   &
            -1.000000000000D+000, -7.387738651055D-001,   &
            -8.896483841425D-001, -6.912196116129D-001,   &
            -6.591363598931D-001, -6.591363598931D-001,   &
            -3.495130204546D-001, -6.431324822055D-001,   &
            -7.354498064526D-003 -6.431324821524D-001,   &
            3.182727192850D-001, -6.591363597049D-001,   &
            5.808679952631D-001, -6.912196111777D-001,   &
            7.387738650537D-001, -7.387738650537D-001,   &
            -1.000000000000D+000, -4.779249498104D-001,   &
            -8.822870241066D-001, -3.979515790090D-001,   &
            -6.431324822055D-001, -3.495130204546D-001,   &
            -3.333333332674D-001, -3.333333332674D-001,   &
            -7.354497586861D-003 -3.495130199382D-001,   &
            2.802386021935D-001, -3.979515779714D-001,   &
            4.779249497027D-001, -4.779249497027D-001,   &
            -1.000000000000D+000, -1.652789576664D-001,   &
            -8.798350575344D-001, -6.008247252297D-002,   &
            -6.431324821524D-001, -7.354498064527D-003,   &
            -3.495130199382D-001, -7.354497586861D-003,   &
            -6.008247113635D-002, -6.008247113635D-002,   &
            1.652789575919D-001, -1.652789575919D-001,   &
            -1.000000000000D+000, 1.652789576664D-001,   &
            -8.822870241144D-001, 2.802386010868D-001,   &
            -6.591363597049D-001, 3.182727192850D-001,   &
            -3.979515779714D-001, 2.802386021935D-001,   &
            -1.652789575919D-001, 1.652789575919D-001,   &
            -1.000000000000D+000, 4.779249498104D-001,   &
            -8.896483841442D-001, 5.808679947612D-001,   &
            -6.912196111777D-001, 5.808679952631D-001,   &
            -4.779249497027D-001, 4.779249497027D-001,   &
            -1.000000000000D+000, 7.387738651055D-001,   &
            -9.021308609041D-001, 8.042617216313D-001,   &
            -7.387738650537D-001, 7.387738650537D-001,   &
            -1.000000000000D+000, 9.195339081665D-001,   &
            -9.195339081537D-001, 9.195339081537D-001    &
            /) , (/2,ND/) )

    CASE (10)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.340014304081D-001, -1.000000000000D+000,   &
            -7.844834736631D-001, -1.000000000000D+000,   &
            -5.652353269962D-001, -1.000000000000D+000,   &
            -2.957581355869D-001, -1.000000000000D+000,   &
            1.147670565004D-016, -1.000000000000D+000,   &
            2.957581355869D-001, -1.000000000000D+000,   &
            5.652353269962D-001, -1.000000000000D+000,   &
            7.844834736631D-001, -1.000000000000D+000,   &
            9.340014304081D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.340014304081D-001,   &
            -9.204846034051D-001, -9.204846034051D-001,   &
            -7.467342693428D-001, -9.103190617998D-001,   &
            -4.997446771417D-001, -9.036831126584D-001,   &
            -2.055546247743D-001, -9.003733097470D-001,   &
            1.059279315451D-001, -9.003733097472D-001,   &
            4.034277877573D-001, -9.036831126663D-001,   &
            6.570533301790D-001, -9.103190617943D-001,   &
            8.409692064794D-001, -9.204846033264D-001,   &
            9.340014303967D-001, -9.340014303967D-001,   &
            -1.000000000000D+000, -7.844834736631D-001,   &
            -9.103190617998D-001, -7.467342693428D-001,   &
            -7.194660870284D-001, -7.194660870284D-001,   &
            -4.556263523466D-001, -7.031799893522D-001,   &
            -1.511276875123D-001, -6.977446262837D-001,   &
            1.588063406409D-001, -7.031799892848D-001,   &
            4.389321734308D-001, -7.194660868432D-001,   &
            6.570533306515D-001, -7.467342689490D-001,   &
            7.844834736180D-001, -7.844834736180D-001,   &
            -1.000000000000D+000, -5.652353269962D-001,   &
            -9.036831126584D-001, -4.997446771417D-001,   &
            -7.031799893522D-001, -4.556263523466D-001,   &
            -4.335320518231D-001, -4.335320518231D-001,   &
            -1.329358964208D-001, -4.335320516356D-001,   &
            1.588063412535D-001, -4.556263517477D-001,   &
            4.034277888813D-001, -4.997446761588D-001,   &
            5.652353268949D-001, -5.652353268949D-001,   &
            -1.000000000000D+000, -2.957581355869D-001,   &
            -9.003733097470D-001, -2.055546247743D-001,   &
            -6.977446262837D-001, -1.511276875123D-001,   &
            -4.335320516356D-001, -1.329358964208D-001,   &
            -1.511276866437D-001, -1.511276866437D-001,   &
            1.059279331442D-001, -2.055546232261D-001,   &
            2.957581354714D-001, -2.957581354714D-001,   &
            -1.000000000000D+000, -1.876866230406D-017,   &
            -9.003733097472D-001, 1.059279315451D-001,   &
            -7.031799892848D-001, 1.588063406409D-001,   &
            -4.556263517477D-001, 1.588063412535D-001,   &
            -2.055546232261D-001, 1.059279331442D-001,   &
            -4.734044120261D-016, 5.207539594513D-016,   &
            -1.000000000000D+000, 2.957581355869D-001,   &
            -9.036831126663D-001, 4.034277877573D-001,   &
            -7.194660868432D-001, 4.389321734308D-001,   &
            -4.997446761588D-001, 4.034277888813D-001,   &
            -2.957581354714D-001, 2.957581354714D-001,   &
            -1.000000000000D+000, 5.652353269962D-001,   &
            -9.103190617943D-001, 6.570533301790D-001,   &
            -7.467342689490D-001, 6.570533306515D-001,   &
            -5.652353268949D-001, 5.652353268949D-001,   &
            -1.000000000000D+000, 7.844834736631D-001,   &
            -9.204846033264D-001, 8.409692064794D-001,   &
            -7.844834736180D-001, 7.844834736180D-001,   &
            -1.000000000000D+000, 9.340014304081D-001,   &
            -9.340014303967D-001, 9.340014303967D-001    &
            /) , (/2,ND/) )

    CASE (11)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.448992722229D-001, -1.000000000000D+000,   &
            -8.192793216440D-001, -1.000000000000D+000,   &
            -6.328761530319D-001, -1.000000000000D+000,   &
            -3.995309409653D-001, -1.000000000000D+000,   &
            -1.365529328549D-001, -1.000000000000D+000,   &
            1.365529328549D-001, -1.000000000000D+000,   &
            3.995309409653D-001, -1.000000000000D+000,   &
            6.328761530319D-001, -1.000000000000D+000,   &
            8.192793216440D-001, -1.000000000000D+000,   &
            9.448992722229D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.448992722229D-001,   &
            -9.341974119518D-001, -9.341974119518D-001,   &
            -7.888939534691D-001, -9.258585304419D-001,   &
            -5.789602117478D-001, -9.200471336715D-001,   &
            -3.228771300183D-001, -9.165753679316D-001,   &
            -4.229026002680D-002, -9.154194833404D-001,   &
            2.394524950301D-001, -9.165753679288D-001,   &
            4.990073435662D-001, -9.200471336748D-001,   &
            7.147524830480D-001, -9.258585304275D-001,   &
            8.683948236054D-001, -9.341974118808D-001,   &
            9.448992722133D-001, -9.448992722133D-001,   &
            -1.000000000000D+000, -8.192793216440D-001,   &
            -9.258585304419D-001, -7.888939534691D-001,   &
            -7.658638517615D-001, -7.658638517615D-001,   &
            -5.400295419371D-001, -7.506188436020D-001,   &
            -2.713892540078D-001, -7.429923084761D-001,   &
            1.438156070715D-002, -7.429923084557D-001,   &
            2.906483843123D-001, -7.506188435426D-001,   &
            5.317277028459D-001, -7.658638515895D-001,   &
            7.147524834584D-001, -7.888939531377D-001,   &
            8.192793216067D-001, -8.192793216067D-001,   &
            -1.000000000000D+000, -6.328761530319D-001,   &
            -9.200471336715D-001, -5.789602117478D-001,   &
            -7.506188436020D-001, -5.400295419371D-001,   &
            -5.166604300986D-001, -5.166604300986D-001,   &
            -2.455728366475D-001, -5.088543269265D-001,   &
            3.332085986519D-002, -5.166604298428D-001,   &
            2.906483849750D-001, -5.400295413485D-001,   &
            4.990073445944D-001, -5.789602109071D-001,   &
            6.328761529454D-001, -6.328761529454D-001,   &
            -1.000000000000D+000, -3.995309409653D-001,   &
            -9.165753679316D-001, -3.228771300183D-001,   &
            -7.429923084761D-001, -2.713892540078D-001,   &
            -5.088543269265D-001, -2.455728366475D-001,   &
            -2.455728363067D-001, -2.455728363067D-001,   &
            1.438156181626D-002, -2.713892529607D-001,   &
            2.394524966200D-001, -3.228771285643D-001,   &
            3.995309408421D-001, -3.995309408421D-001,   &
            -1.000000000000D+000, -1.365529328549D-001,   &
            -9.154194833404D-001, -4.229026002680D-002,   &
            -7.429923084557D-001, 1.438156070715D-002,   &
            -5.166604298428D-001, 3.332085986519D-002,   &
            -2.713892529607D-001, 1.438156181626D-002,   &
            -4.229025824416D-002, -4.229025824416D-002,   &
            1.365529327886D-001, -1.365529327886D-001,   &
            -1.000000000000D+000, 1.365529328549D-001,   &
            -9.165753679288D-001, 2.394524950301D-001,   &
            -7.506188435426D-001, 2.906483843123D-001,   &
            -5.400295413485D-001, 2.906483849750D-001,   &
            -3.228771285643D-001, 2.394524966200D-001,   &
            -1.365529327886D-001, 1.365529327886D-001,   &
            -1.000000000000D+000, 3.995309409653D-001,   &
            -9.200471336748D-001, 4.990073435662D-001,   &
            -7.658638515895D-001, 5.317277028459D-001,   &
            -5.789602109071D-001, 4.990073445944D-001,   &
            -3.995309408421D-001, 3.995309408421D-001,   &
            -1.000000000000D+000, 6.328761530319D-001,   &
            -9.258585304275D-001, 7.147524830480D-001,   &
            -7.888939531377D-001, 7.147524834584D-001,   &
            -6.328761529454D-001, 6.328761529454D-001,   &
            -1.000000000000D+000, 8.192793216440D-001,   &
            -9.341974118808D-001, 8.683948236054D-001,   &
            -8.192793216067D-001, 8.192793216067D-001,   &
            -1.000000000000D+000, 9.448992722229D-001,   &
            -9.448992722133D-001, 9.448992722133D-001    &
            /) , (/2,ND/) )

    CASE (12)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.533098466422D-001, -1.000000000000D+000,   &
            -8.463475646519D-001, -1.000000000000D+000,   &
            -6.861884690818D-001, -1.000000000000D+000,   &
            -4.829098210913D-001, -1.000000000000D+000,   &
            -2.492869301062D-001, -1.000000000000D+000,   &
            1.203298976554D-016, -1.000000000000D+000,   &
            2.492869301062D-001, -1.000000000000D+000,   &
            4.829098210913D-001, -1.000000000000D+000,   &
            6.861884690818D-001, -1.000000000000D+000,   &
            8.463475646519D-001, -1.000000000000D+000,   &
            9.533098466422D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.533098466422D-001,   &
            -9.446940110119D-001, -9.446940110119D-001,   &
            -8.215657972316D-001, -9.377923030321D-001,   &
            -6.414618243036D-001, -9.327587160826D-001,   &
            -4.178368260535D-001, -9.294242521054D-001,   &
            -1.666672186572D-001, -9.277601017008D-001,   &
            9.442731687147D-002, -9.277601016950D-001,   &
            3.472610754367D-001, -9.294242520974D-001,   &
            5.742205387344D-001, -9.327587160786D-001,   &
            7.593580994944D-001, -9.377923030093D-001,   &
            8.893880217562D-001, -9.446940109476D-001,   &
            9.533098466340D-001, -9.533098466340D-001,   &
            -1.000000000000D+000, -8.463475646519D-001,   &
            -9.377923030321D-001, -8.215657972316D-001,   &
            -8.020988549335D-001, -8.020988549335D-001,   &
            -6.075481604223D-001, -7.883274738251D-001,   &
            -3.709518388955D-001, -7.800690331410D-001,   &
            -1.113423193204D-001, -7.773153637825D-001,   &
            1.510208699587D-001, -7.800690331184D-001,   &
            3.958756329335D-001, -7.883274737731D-001,   &
            6.041977091655D-001, -8.020988547712D-001,   &
            7.593580998476D-001, -8.215657969514D-001,   &
            8.463475646204D-001, -8.463475646204D-001,   &
            -1.000000000000D+000, -6.861884690818D-001,   &
            -9.327587160826D-001, -6.414618243036D-001,   &
            -7.883274738251D-001, -6.075481604223D-001,   &
            -5.849629322706D-001, -5.849629322706D-001,   &
            -3.427539956851D-001, -5.736474873433D-001,   &
            -8.359851756886D-002, -5.736474872566D-001,   &
            1.699258639917D-001, -5.849629320013D-001,   &
            3.958756336084D-001, -6.075481598680D-001,   &
            5.742205396532D-001, -6.414618235925D-001,   &
            6.861884690079D-001, -6.861884690079D-001,   &
            -1.000000000000D+000, -4.829098210913D-001,   &
            -9.294242521054D-001, -4.178368260535D-001,   &
            -7.800690331410D-001, -3.709518388955D-001,   &
            -5.736474873433D-001, -3.427539956851D-001,   &
            -3.333333332340D-001, -3.333333332340D-001,   &
            -8.359851708137D-002, -3.427539951644D-001,   &
            1.510208712232D-001, -3.709518377970D-001,   &
            3.472610769425D-001, -4.178368247579D-001,   &
            4.829098209730D-001, -4.829098209730D-001,   &
            -1.000000000000D+000, -2.492869301062D-001,   &
            -9.277601017008D-001, -1.666672186572D-001,   &
            -7.773153637825D-001, -1.113423193204D-001,   &
            -5.736474872566D-001, -8.359851756886D-002,   &
            -3.427539951644D-001, -8.359851708137D-002,   &
            -1.113423178906D-001, -1.113423178906D-001,   &
            9.442731872758D-002, -1.666672168825D-001,   &
            2.492869300008D-001, -2.492869300008D-001,   &
            -1.000000000000D+000, -8.880216002865D-018,   &
            -9.277601016950D-001, 9.442731687147D-002,   &
            -7.800690331184D-001, 1.510208699587D-001,   &
            -5.849629320013D-001, 1.699258639917D-001,   &
            -3.709518377970D-001, 1.510208712232D-001,   &
            -1.666672168825D-001, 9.442731872758D-002,   &
            4.533477103045D-016, -4.575332210522D-016,   &
            -1.000000000000D+000, 2.492869301062D-001,   &
            -9.294242520974D-001, 3.472610754367D-001,   &
            -7.883274737731D-001, 3.958756329335D-001,   &
            -6.075481598680D-001, 3.958756336084D-001,   &
            -4.178368247579D-001, 3.472610769425D-001,   &
            -2.492869300008D-001, 2.492869300008D-001,   &
            -1.000000000000D+000, 4.829098210913D-001,   &
            -9.327587160786D-001, 5.742205387344D-001,   &
            -8.020988547712D-001, 6.041977091655D-001,   &
            -6.414618235925D-001, 5.742205396532D-001,   &
            -4.829098209730D-001, 4.829098209730D-001,   &
            -1.000000000000D+000, 6.861884690818D-001,   &
            -9.377923030093D-001, 7.593580994944D-001,   &
            -8.215657969514D-001, 7.593580998476D-001,   &
            -6.861884690079D-001, 6.861884690079D-001,   &
            -1.000000000000D+000, 8.463475646519D-001,   &
            -9.446940109476D-001, 8.893880217562D-001,   &
            -8.463475646204D-001, 8.463475646204D-001,   &
            -1.000000000000D+000, 9.533098466422D-001,   &
            -9.533098466340D-001, 9.533098466340D-001    &
            /) , (/2,ND/) )

    CASE (13)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.599350452673D-001, -1.000000000000D+000,   &
            -8.678010538303D-001, -1.000000000000D+000,   &
            -7.288685990913D-001, -1.000000000000D+000,   &
            -5.506394029286D-001, -1.000000000000D+000,   &
            -3.427240133427D-001, -1.000000000000D+000,   &
            -1.163318688837D-001, -1.000000000000D+000,   &
            1.163318688837D-001, -1.000000000000D+000,   &
            3.427240133427D-001, -1.000000000000D+000,   &
            5.506394029286D-001, -1.000000000000D+000,   &
            7.288685990913D-001, -1.000000000000D+000,   &
            8.678010538303D-001, -1.000000000000D+000,   &
            9.599350452673D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.599350452673D-001,   &
            -9.528966352645D-001, -9.528966352645D-001,   &
            -8.473433430483D-001, -9.471313846225D-001,   &
            -6.914534709425D-001, -9.427840249816D-001,   &
            -4.952450730697D-001, -9.397047663002D-001,   &
            -2.707301857968D-001, -9.378624472876D-001,   &
            -3.137576336546D-002, -9.372484771762D-001,   &
            2.085926295549D-001, -9.378624472728D-001,   &
            4.349498367823D-001, -9.397047662850D-001,   &
            6.342374943925D-001, -9.427840249684D-001,   &
            7.944747269525D-001, -9.471313845907D-001,   &
            9.057932702777D-001, -9.528966352032D-001,   &
            9.599350452597D-001, -9.599350452597D-001,   &
            -1.000000000000D+000, -8.678010538303D-001,   &
            -9.471313846225D-001, -8.473433430483D-001,   &
            -8.308213623361D-001, -8.308213623361D-001,   &
            -6.620335860676D-001, -8.185797272634D-001,   &
            -4.533080802956D-001, -8.104335611591D-001,   &
            -2.189923442145D-001, -8.063605149473D-001,   &
            2.535285612122D-002, -8.063605149379D-001,   &
            2.637416391093D-001, -8.104335611378D-001,   &
            4.806133119176D-001, -8.185797272100D-001,   &
            6.616427239281D-001, -8.308213621716D-001,   &
            7.944747272690D-001, -8.473433427976D-001,   &
            8.678010538017D-001, -8.678010538017D-001,   &
            -1.000000000000D+000, -7.288685990913D-001,   &
            -9.427840249816D-001, -6.914534709425D-001,   &
            -8.185797272634D-001, -6.620335860676D-001,   &
            -6.410774090477D-001, -6.410774090477D-001,   &
            -4.253068029291D-001, -6.284845814082D-001,   &
            -1.878588344746D-001, -6.242823321166D-001,   &
            5.379138335949D-002, -6.284845812830D-001,   &
            2.821548173387D-001, -6.410774087703D-001,   &
            4.806133126083D-001, -6.620335855263D-001,   &
            6.342374952401D-001, -6.914534703123D-001,   &
            7.288685990238D-001, -7.288685990238D-001,   &
            -1.000000000000D+000, -5.506394029286D-001,   &
            -9.397047663002D-001, -4.952450730697D-001,   &
            -8.104335611591D-001, -4.533080802956D-001,   &
            -6.284845814082D-001, -4.253068029291D-001,   &
            -4.112704522877D-001, -4.112704522877D-001,   &
            -1.774590953869D-001, -4.112704520842D-001,   &
            5.379138399500D-002, -4.253068022998D-001,   &
            2.637416405137D-001, -4.533080791632D-001,   &
            4.349498382308D-001, -4.952450718885D-001,   &
            5.506394028132D-001, -5.506394028132D-001,   &
            -1.000000000000D+000, -3.427240133427D-001,   &
            -9.378624472876D-001, -2.707301857968D-001,   &
            -8.063605149473D-001, -2.189923442145D-001,   &
            -6.242823321166D-001, -1.878588344746D-001,   &
            -4.112704520842D-001, -1.774590953869D-001,   &
            -1.878588336319D-001, -1.878588336319D-001,   &
            2.535285791130D-002, -2.189923425490D-001,   &
            2.085926314627D-001, -2.707301840606D-001,   &
            3.427240132123D-001, -3.427240132123D-001,   &
            -1.000000000000D+000, -1.163318688837D-001,   &
            -9.372484771762D-001, -3.137576336546D-002,   &
            -8.063605149379D-001, 2.535285612122D-002,   &
            -6.284845812830D-001, 5.379138335949D-002,   &
            -4.253068022998D-001, 5.379138399500D-002,   &
            -2.189923425490D-001, 2.535285791130D-002,   &
            -3.137576133722D-002, -3.137576133722D-002,   &
            1.163318688227D-001, -1.163318688227D-001,   &
            -1.000000000000D+000, 1.163318688837D-001,   &
            -9.378624472728D-001, 2.085926295549D-001,   &
            -8.104335611378D-001, 2.637416391093D-001,   &
            -6.410774087703D-001, 2.821548173387D-001,   &
            -4.533080791632D-001, 2.637416405137D-001,   &
            -2.707301840606D-001, 2.085926314627D-001,   &
            -1.163318688227D-001, 1.163318688227D-001,   &
            -1.000000000000D+000, 3.427240133427D-001,   &
            -9.397047662850D-001, 4.349498367823D-001,   &
            -8.185797272100D-001, 4.806133119176D-001,   &
            -6.620335855263D-001, 4.806133126083D-001,   &
            -4.952450718885D-001, 4.349498382308D-001,   &
            -3.427240132123D-001, 3.427240132123D-001,   &
            -1.000000000000D+000, 5.506394029286D-001,   &
            -9.427840249684D-001, 6.342374943925D-001,   &
            -8.308213621716D-001, 6.616427239281D-001,   &
            -6.914534703123D-001, 6.342374952401D-001,   &
            -5.506394028132D-001, 5.506394028132D-001,   &
            -1.000000000000D+000, 7.288685990913D-001,   &
            -9.471313845907D-001, 7.944747269525D-001,   &
            -8.473433427976D-001, 7.944747272690D-001,   &
            -7.288685990238D-001, 7.288685990238D-001,   &
            -1.000000000000D+000, 8.678010538303D-001,   &
            -9.528966352032D-001, 9.057932702777D-001,   &
            -8.678010538017D-001, 8.678010538017D-001,   &
            -1.000000000000D+000, 9.599350452673D-001,   &
            -9.599350452597D-001, 9.599350452597D-001    &
            /) , (/2,ND/) )

    CASE (14)
       p2D(:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000D+000, -1.000000000000D+000,   &
            1.000000000000D+000, -1.000000000000D+000,   &
            -1.000000000000D+000, 1.000000000000D+000,   &
            -9.652459265038D-001, -1.000000000000D+000,   &
            -8.850820442230D-001, -1.000000000000D+000,   &
            -7.635196899518D-001, -1.000000000000D+000,   &
            -6.062532054698D-001, -1.000000000000D+000,   &
            -4.206380547137D-001, -1.000000000000D+000,   &
            -2.153539553638D-001, -1.000000000000D+000,   &
            1.105861828404D-016, -1.000000000000D+000,   &
            2.153539553638D-001, -1.000000000000D+000,   &
            4.206380547137D-001, -1.000000000000D+000,   &
            6.062532054698D-001, -1.000000000000D+000,   &
            7.635196899518D-001, -1.000000000000D+000,   &
            8.850820442230D-001, -1.000000000000D+000,   &
            9.652459265038D-001, -1.000000000000D+000,   &
            -1.000000000000D+000, -9.652459265038D-001,   &
            -9.594219087433D-001, -9.594219087433D-001,   &
            -8.680068205242D-001, -9.545625190217D-001,   &
            -7.319575708858D-001, -9.508037148963D-001,   &
            -5.588829169561D-001, -9.480138853402D-001,   &
            -3.579694504810D-001, -9.461613006493D-001,   &
            -1.396449437646D-001, -9.452357419067D-001,   &
            8.488068160544D-002, -9.452357418956D-001,   &
            3.041307477122D-001, -9.461613006250D-001,   &
            5.068967998913D-001, -9.480138853174D-001,   &
            6.827612843767D-001, -9.508037148740D-001,   &
            8.225693388804D-001, -9.545625189824D-001,   &
            9.188438172526D-001, -9.594219086853D-001,   &
            9.652459264968D-001, -9.652459264968D-001,   &
            -1.000000000000D+000, -8.850820442230D-001,   &
            -9.545625190217D-001, -8.680068205242D-001,   &
            -8.539072359035D-001, -8.539072359035D-001,   &
            -7.064321386476D-001, -8.430994429661D-001,   &
            -5.216752587544D-001, -8.354046456073D-001,   &
            -3.106270658509D-001, -8.307906004989D-001,   &
            -8.537395985816D-002, -8.292520841395D-001,   &
            1.414176628890D-001, -8.307906004852D-001,   &
            3.570799018687D-001, -8.354046455868D-001,   &
            5.495315801493D-001, -8.430994429072D-001,   &
            7.078144710405D-001, -8.539072357376D-001,   &
            8.225693391615D-001, -8.680068203003D-001,   &
            8.850820441967D-001, -8.850820441967D-001,   &
            -1.000000000000D+000, -7.635196899518D-001,   &
            -9.508037148963D-001, -7.319575708858D-001,   &
            -8.430994429661D-001, -7.064321386476D-001,   &
            -6.873749345785D-001, -6.873749345785D-001,   &
            -4.950855938967D-001, -6.746606983487D-001,   &
            -2.789847461918D-001, -6.682966869954D-001,   &
            -5.271856842992D-002, -6.682966869499D-001,   &
            1.697462909436D-001, -6.746606982125D-001,   &
            3.747498682323D-001, -6.873749343015D-001,   &
            5.495315808319D-001, -7.064321381277D-001,   &
            6.827612851476D-001, -7.319575703308D-001,   &
            7.635196898900D-001, -7.635196898900D-001,   &
            -1.000000000000D+000, -6.062532054698D-001,   &
            -9.480138853402D-001, -5.588829169561D-001,   &
            -8.354046456073D-001, -5.216752587544D-001,   &
            -6.746606983487D-001, -4.950855938967D-001,   &
            -4.790930768063D-001, -4.790930768063D-001,   &
            -2.631227288402D-001, -4.737545424098D-001,   &
            -4.181384665401D-002, -4.790930764888D-001,   &
            1.697462916912D-001, -4.950855932180D-001,   &
            3.570799033414D-001, -5.216752576392D-001,   &
            5.068968012493D-001, -5.588829159004D-001,   &
            6.062532053602D-001, -6.062532053602D-001,   &
            -1.000000000000D+000, -4.206380547137D-001,   &
            -9.461613006493D-001, -3.579694504810D-001,   &
            -8.307906004989D-001, -3.106270658509D-001,   &
            -6.682966869954D-001, -2.789847461918D-001,   &
            -4.737545424098D-001, -2.631227288402D-001,   &
            -2.631227284887D-001, -2.631227284887D-001,   &
            -5.271856728199D-002, -2.789847451219D-001,   &
            1.414176649318D-001, -3.106270640805D-001,   &
            3.041307495879D-001, -3.579694488568D-001,   &
            4.206380545720D-001, -4.206380545720D-001,   &
            -1.000000000000D+000, -2.153539553638D-001,   &
            -9.452357419067D-001, -1.396449437646D-001,   &
            -8.292520841395D-001, -8.537395985816D-002,   &
            -6.682966869499D-001, -5.271856842992D-002,   &
            -4.790930764888D-001, -4.181384665401D-002,   &
            -2.789847451219D-001, -5.271856728199D-002,   &
            -8.537395770354D-002, -8.537395770354D-002,   &
            8.488068375095D-002, -1.396449417133D-001,   &
            2.153539552554D-001, -2.153539552554D-001,   &
            -1.000000000000D+000, -3.518772397373D-017,   &
            -9.452357418956D-001, 8.488068160544D-002,   &
            -8.307906004852D-001, 1.414176628890D-001,   &
            -6.746606982125D-001, 1.697462909436D-001,   &
            -4.950855932180D-001, 1.697462916912D-001,   &
            -3.106270640805D-001, 1.414176649318D-001,   &
            -1.396449417133D-001, 8.488068375095D-002,   &
            -9.122915050063D-016, 9.054332831098D-016,   &
            -1.000000000000D+000, 2.153539553638D-001,   &
            -9.461613006250D-001, 3.041307477122D-001,   &
            -8.354046455868D-001, 3.570799018687D-001,   &
            -6.873749343015D-001, 3.747498682323D-001,   &
            -5.216752576392D-001, 3.570799033414D-001,   &
            -3.579694488568D-001, 3.041307495879D-001,   &
            -2.153539552554D-001, 2.153539552554D-001,   &
            -1.000000000000D+000, 4.206380547137D-001,   &
            -9.480138853174D-001, 5.068967998913D-001,   &
            -8.430994429072D-001, 5.495315801493D-001,   &
            -7.064321381277D-001, 5.495315808319D-001,   &
            -5.588829159004D-001, 5.068968012493D-001,   &
            -4.206380545720D-001, 4.206380545720D-001,   &
            -1.000000000000D+000, 6.062532054698D-001,   &
            -9.508037148740D-001, 6.827612843767D-001,   &
            -8.539072357376D-001, 7.078144710405D-001,   &
            -7.319575703308D-001, 6.827612851476D-001,   &
            -6.062532053602D-001, 6.062532053602D-001,   &
            -1.000000000000D+000, 7.635196899518D-001,   &
            -9.545625189824D-001, 8.225693388804D-001,   &
            -8.680068203003D-001, 8.225693391615D-001,   &
            -7.635196898900D-001, 7.635196898900D-001,   &
            -1.000000000000D+000, 8.850820442230D-001,   &
            -9.594219086853D-001, 9.188438172526D-001,   &
            -8.850820441967D-001, 8.850820441967D-001,   &
            -1.000000000000D+000, 9.652459265038D-001,   &
            -9.652459264968D-001, 9.652459264968D-001    &
            /) , (/2,ND/) )

    CASE DEFAULT

       STOP 'HighOrderCell_feketeNodes2D: not implemented'

    END SELECT

  END SUBROUTINE HighOrderCell_feketeNodes2D


  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  SUBROUTINE HighOrderCell_feketeNodes3D(p, p3D)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    REAL*8,  INTENT(out) :: p3D(3,*)
    INTEGER :: ND

    ND = (p+1)*(p+2)*(p+3)/6

    SELECT CASE(p)
    CASE(1)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1,-1,-1,    &
            1,-1,-1,    &
            -1,1,-1,    &
            -1,-1,1     &
            /) , (/3,ND/) )

    CASE(2)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1,-1,-1,    &
            1,-1,-1,    &
            -1,1,-1,    &
            -1,-1,1,    &
            0,-1,-1,    &
            -1,0,-1,    &
            0,0,-1,    &
            -1,-1,0,    &
            0,-1,0,    &
            -1,0,0     &
            /) , (/3,ND/) )

    CASE(3)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,1.000000000000000,    &
            -0.447213595499958,-1.000000000000000,-1.000000000000000,    &
            0.447213595499958,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-0.447213595499958,-1.000000000000000,    &
            -0.333333333333333,-0.333333333333333,-1.000000000000000,    &
            0.447213595499958,-0.447213595499958,-1.000000000000000,    &
            -1.000000000000000,0.447213595499958,-1.000000000000000,    &
            -0.447213595499958,0.447213595499958,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,-0.447213595499958,    &
            -0.333333333333333,-1.000000000000000,-0.333333333333333,    &
            0.447213595499958,-1.000000000000000,-0.447213595499958,    &
            -1.000000000000000,-0.333333333333333,-0.333333333333333,    &
            -0.333333333333333,-0.333333333333333,-0.333333333333333,    &
            -1.000000000000000,0.447213595499958,-0.447213595499958,    &
            -1.000000000000000,-1.000000000000000,0.447213595499958,    &
            -0.447213595499958,-1.000000000000000,0.447213595499958,    &
            -1.000000000000000,-0.447213595499958,0.447213595499958     &
            /) , (/3,ND/) )

    CASE(4)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,1.000000000000000,    &
            -0.654653670707977,-1.000000000000000,-1.000000000000000,    &
            0.0,-1.000000000000000,-1.000000000000000,    &
            0.654653670707977,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-0.654653670707977,-1.000000000000000,    &
            -0.551551223569326,-0.551551223569326,-1.000000000000000,    &
            0.103102447138651,-0.551551223569326,-1.000000000000000,    &
            0.654653670707977,-0.654653670707977,-1.000000000000000,    &
            -1.000000000000000           ,0.0,-1.000000000000000,    &
            -0.551551223569326,0.103102447138651,-1.000000000000000,    &
            0.0           ,0.0,-1.000000000000000,    &
            -1.000000000000000,0.654653670707977,-1.000000000000000,    &
            -0.654653670707977,0.654653670707977,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,-0.654653670707977,    &
            -0.551551223569326,-1.000000000000000,-0.551551223569326,    &
            0.103102447138651,-1.000000000000000,-0.551551223569326,    &
            0.654653670707977,-1.000000000000000,-0.654653670707977,    &
            -1.000000000000000,-0.551551223569326,-0.551551223569326,    &
            -0.500000000000000,-0.500000000000000,-0.500000000000000,    &
            0.103102447138651,-0.551551223569326,-0.551551223569326,    &
            -1.000000000000000,0.103102447138651,-0.551551223569326,    &
            -0.551551223569326,0.103102447138651,-0.551551223569326,    &
            -1.000000000000000,0.654653670707977,-0.654653670707977,    &
            -1.000000000000000,-1.000000000000000           ,0.0,    &
            -0.551551223569326,-1.000000000000000,0.103102447138651,    &
            0.0,-1.000000000000000           ,0.0,    &
            -1.000000000000000,-0.551551223569326,0.103102447138651,    &
            -0.551551223569326,-0.551551223569326,0.103102447138651,    &
            -1.000000000000000           ,0.0           ,0.0,    &
            -1.000000000000000,-1.000000000000000,0.654653670707977,    &
            -0.654653670707977,-1.000000000000000,0.654653670707977,    &
            -1.000000000000000,-0.654653670707977,0.654653670707977     &
            /) , (/3,ND/) )

    CASE(5)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            1.000000000000000,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,1.000000000000000,    &
            -0.765055323929465,-1.000000000000000,-1.000000000000000,    &
            -0.285231516480645,-1.000000000000000,-1.000000000000000,    &
            0.285231516480645,-1.000000000000000,-1.000000000000000,    &
            0.765055323929465,-1.000000000000000,-1.000000000000000,    &
            -1.000000000000000,-0.765055323929465,-1.000000000000000,    &
            -0.683428946803370,-0.683428946803370,-1.000000000000000,    &
            -0.173392064183727,-0.653215871632546,-1.000000000000000,    &
            0.366857893606740,-0.683428946803370,-1.000000000000000,    &
            0.765055323929465,-0.765055323929465,-1.000000000000000,    &
            -1.000000000000000,-0.285231516480645,-1.000000000000000,    &
            -0.653215871632546,-0.173392064183727,-1.000000000000000,    &
            -0.173392064183727,-0.173392064183727,-1.000000000000000,    &
            0.285231516480645,-0.285231516480645,-1.000000000000000,    &
            -1.000000000000000,0.285231516480645,-1.000000000000000,    &
            -0.683428946803370,0.366857893606740,-1.000000000000000,    &
            -0.285231516480645,0.285231516480645,-1.000000000000000,    &
            -1.000000000000000,0.765055323929465,-1.000000000000000,    &
            -0.765055323929465,0.765055323929465,-1.000000000000000,    &
            -1.000000000000000,-1.000000000000000,-0.765055323929465,    &
            -0.683428946803370,-1.000000000000000,-0.683428946803370,    &
            -0.173392064183727,-1.000000000000000,-0.653215871632546,    &
            0.366857893606740,-1.000000000000000,-0.683428946803370,    &
            0.765055323929465,-1.000000000000000,-0.765055323929465,    &
            -1.000000000000000,-0.683428946803370,-0.683428946803370,    &
            -0.619955951862205,-0.619955951862205,-0.619955951862205,    &
            -0.140132144413385,-0.619955951862205,-0.619955951862205,    &
            0.366857893606740,-0.683428946803370,-0.683428946803370,    &
            -1.000000000000000,-0.173392064183727,-0.653215871632546,    &
            -0.619955951862205,-0.140132144413385,-0.619955951862205,    &
            -0.173392064183727,-0.173392064183727,-0.653215871632546,    &
            -1.000000000000000,0.366857893606740,-0.683428946803370,    &
            -0.683428946803370,0.366857893606740,-0.683428946803370,    &
            -1.000000000000000,0.765055323929465,-0.765055323929465,    &
            -1.000000000000000,-1.000000000000000,-0.285231516480645,    &
            -0.653215871632546,-1.000000000000000,-0.173392064183727,    &
            -0.173392064183727,-1.000000000000000,-0.173392064183727,    &
            0.285231516480645,-1.000000000000000,-0.285231516480645,    &
            -1.000000000000000,-0.653215871632546,-0.173392064183727,    &
            -0.619955951862205,-0.619955951862205,-0.140132144413385,    &
            -0.173392064183727,-0.653215871632546,-0.173392064183727,    &
            -1.000000000000000,-0.173392064183727,-0.173392064183727,    &
            -0.653215871632546,-0.173392064183727,-0.173392064183727,    &
            -1.000000000000000,0.285231516480645,-0.285231516480645,    &
            -1.000000000000000,-1.000000000000000,0.285231516480645,    &
            -0.683428946803370,-1.000000000000000,0.366857893606740,    &
            -0.285231516480645,-1.000000000000000,0.285231516480645,    &
            -1.000000000000000,-0.683428946803370,0.366857893606740,    &
            -0.683428946803370,-0.683428946803370,0.366857893606740,    &
            -1.000000000000000,-0.285231516480645,0.285231516480645,    &
            -1.000000000000000,-1.000000000000000,0.765055323929465,    &
            -0.765055323929465,-1.000000000000000,0.765055323929465,    &
            -1.000000000000000,-0.765055323929465,0.765055323929465     &
            /) , (/3,ND/) )

    CASE(6)
       p3D (:,1:ND) = RESHAPE( (/                                 &
            -1.000000000000000e+000,-1.000000000000000e+000,-1.000000000000000e+000,    &
            1.000000000000000e+000,-1.000000000000000e+000,-1.000000000000000e+000,    &
            -1.000000000000000e+000,1.000000000000000e+000,-1.000000000000000e+000,    &
            -1.000000000000000e+000,-1.000000000000000e+000,1.000000000000000e+000,    &
            -8.302238962785670e-001,-1.000000000000000e+000,-1.000000000000000e+000,    &
            -4.688487934707142e-001,-1.000000000000000e+000,-1.000000000000000e+000,    &
            0.0,-1.000000000000000e+000,-1.000000000000000e+000,    &
            4.688487934707142e-001,-1.000000000000000e+000,-1.000000000000000e+000,    &
            8.302238962785671e-001,-1.000000000000000e+000,-1.000000000000000e+000,    &
            -1.000000000000000e+000,-8.302238962785671e-001,-1.000000000000000e+000,    &
            -7.663575632497603e-001,-7.663575632497603e-001,-1.000000000000000e+000,    &
            -3.691578968876205e-001,-7.305329996954733e-001,-1.000000000000000e+000,    &
            9.969089658309382e-002,-7.305329996954733e-001,-1.000000000000000e+000,    &
            5.327151264995207e-001,-7.663575632497605e-001,-1.000000000000000e+000,    &
            8.302238962785671e-001,-8.302238962785670e-001,-1.000000000000000e+000,    &
            -1.000000000000000e+000,-4.688487934707142e-001,-1.000000000000000e+000,    &
            -7.305329996954733e-001,-3.691578968876205e-001,-1.000000000000000e+000,    &
            -3.333333333333334e-001,-3.333333333333334e-001,-1.000000000000000e+000,    &
            9.969089658309382e-002,-3.691578968876205e-001,-1.000000000000000e+000,    &
            4.688487934707142e-001,-4.688487934707142e-001,-1.000000000000000e+000,    &
            -1.000000000000000e+000                      ,0.0,-1.000000000000000e+000,    &
            -7.305329996954733e-001,9.969089658309382e-002,-1.000000000000000e+000,    &
            -3.691578968876205e-001,9.969089658309382e-002,-1.000000000000000e+000,    &
            0.0                      ,0.0,-1.000000000000000e+000,    &
            -1.000000000000000e+000,4.688487934707142e-001,-1.000000000000000e+000,    &
            -7.663575632497605e-001,5.327151264995207e-001,-1.000000000000000e+000,    &
            -4.688487934707142e-001,4.688487934707142e-001,-1.000000000000000e+000,    &
            -1.000000000000000e+000,8.302238962785671e-001,-1.000000000000000e+000,    &
            -8.302238962785671e-001,8.302238962785671e-001,-1.000000000000000e+000,    &
            -1.000000000000000e+000,-1.000000000000000e+000,-8.302238962785670e-001,    &
            -7.663575632497603e-001,-1.000000000000000e+000,-7.663575632497603e-001,    &
            -3.691578968876205e-001,-1.000000000000000e+000,-7.305329996954733e-001,    &
            9.969089658309382e-002,-1.000000000000000e+000,-7.305329996954733e-001,    &
            5.327151264995207e-001,-1.000000000000000e+000,-7.663575632497605e-001,    &
            8.302238962785671e-001,-1.000000000000000e+000,-8.302238962785670e-001,    &
            -1.000000000000000e+000,-7.663575632497603e-001,-7.663575632497603e-001,    &
            -7.075559740696417e-001,-7.075559740696417e-001,-7.075559740696417e-001,    &
            -3.193124485960736e-001,-6.806875514039265e-001,-6.806875514039265e-001,    &
            1.226679222089253e-001,-7.075559740696417e-001,-7.075559740696417e-001,    &
            5.327151264995207e-001,-7.663575632497603e-001,-7.663575632497603e-001,    &
            -1.000000000000000e+000,-3.691578968876205e-001,-7.305329996954733e-001,    &
            -6.806875514039265e-001,-3.193124485960736e-001,-6.806875514039263e-001,    &
            -3.193124485960734e-001,-3.193124485960734e-001,-6.806875514039263e-001,    &
            9.969089658309382e-002,-3.691578968876205e-001,-7.305329996954733e-001,    &
            -1.000000000000000e+000,9.969089658309382e-002,-7.305329996954733e-001,    &
            -7.075559740696417e-001,1.226679222089253e-001,-7.075559740696417e-001,    &
            -3.691578968876206e-001,9.969089658309382e-002,-7.305329996954733e-001,    &
            -1.000000000000000e+000,5.327151264995207e-001,-7.663575632497605e-001,    &
            -7.663575632497602e-001,5.327151264995207e-001,-7.663575632497605e-001,    &
            -1.000000000000000e+000,8.302238962785671e-001,-8.302238962785670e-001,    &
            -1.000000000000000e+000,-1.000000000000000e+000,-4.688487934707142e-001,    &
            -7.305329996954733e-001,-1.000000000000000e+000,-3.691578968876205e-001,    &
            -3.333333333333334e-001,-1.000000000000000e+000,-3.333333333333334e-001,    &
            9.969089658309382e-002,-1.000000000000000e+000,-3.691578968876205e-001,    &
            4.688487934707142e-001,-1.000000000000000e+000,-4.688487934707142e-001,    &
            -1.000000000000000e+000,-7.305329996954733e-001,-3.691578968876205e-001,    &
            -6.806875514039263e-001,-6.806875514039263e-001,-3.193124485960736e-001,    &
            -3.193124485960734e-001,-6.806875514039263e-001,-3.193124485960734e-001,    &
            9.969089658309382e-002,-7.305329996954733e-001,-3.691578968876205e-001,    &
            -1.000000000000000e+000,-3.333333333333334e-001,-3.333333333333334e-001,    &
            -6.806875514039263e-001,-3.193124485960734e-001,-3.193124485960734e-001,    &
            -3.333333333333332e-001,-3.333333333333334e-001,-3.333333333333334e-001,    &
            -1.000000000000000e+000,9.969089658309382e-002,-3.691578968876205e-001,    &
            -7.305329996954734e-001,9.969089658309382e-002,-3.691578968876205e-001,    &
            -1.000000000000000e+000,4.688487934707142e-001,-4.688487934707142e-001,    &
            -1.000000000000000e+000,-1.000000000000000e+000,0.0,    &
            -7.305329996954733e-001,-1.000000000000000e+000,9.969089658309382e-002,    &
            -3.691578968876205e-001,-1.000000000000000e+000,9.969089658309382e-002,    &
            0.0,-1.000000000000000e+000                      ,0.0,    &
            -1.000000000000000e+000,-7.305329996954733e-001,9.969089658309382e-002,    &
            -7.075559740696417e-001,-7.075559740696417e-001,1.226679222089253e-001,    &
            -3.691578968876206e-001,-7.305329996954733e-001,9.969089658309382e-002,    &
            -1.000000000000000e+000,-3.691578968876205e-001,9.969089658309382e-002,    &
            -7.305329996954733e-001,-3.691578968876205e-001,9.969089658309382e-002,    &
            -1.000000000000000e+000                      ,0.0                ,0.0,    &
            -1.000000000000000e+000,-1.000000000000000e+000,4.688487934707142e-001,    &
            -7.663575632497605e-001,-1.000000000000000e+000,5.327151264995207e-001,    &
            -4.688487934707142e-001,-1.000000000000000e+000,4.688487934707142e-001,    &
            -1.000000000000000e+000,-7.663575632497605e-001,5.327151264995207e-001,    &
            -7.663575632497603e-001,-7.663575632497605e-001,5.327151264995207e-001,    &
            -1.000000000000000e+000,-4.688487934707142e-001,4.688487934707142e-001,    &
            -1.000000000000000e+000,-1.000000000000000e+000,8.302238962785671e-001,    &
            -8.302238962785671e-001,-1.000000000000000e+000,8.302238962785671e-001,    &
            -1.000000000000000e+000,-8.302238962785671e-001,8.302238962785671e-001     &
            /) , (/3,ND/) )
    CASE DEFAULT

       STOP 'HighOrderCell_feketeNodes3D: not implemented'
    END SELECT

  END SUBROUTINE HighOrderCell_feketeNodes3D


  SUBROUTINE  HighOrderCell_SetNodes1D(p, optionNodes, coordinates)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    REAL*8, INTENT(out)  :: coordinates(*)
    INTEGER :: i
    REAL*8  :: h0
    IF (optionNodes==0) THEN
       coordinates(1) = -1
       coordinates(2) =  1
       h0 = 2.d0 / p
       DO i = 3, p+1
          coordinates(i) = -1 +(i-2)* h0
       ENDDO
    ELSE IF( optionNodes==1)THEN
       CALL HighOrderCell_feketeNodes1D(p, coordinates)
    END IF
  END SUBROUTINE HighOrderCell_SetNodes1D



  SUBROUTINE  HighOrderCell_SetNodesTRI(p, optionNodes, nodes)
    !
    ! nodes = HighOrderCell_SetNodesTRI(p, optionNodes)
    !
    ! p:            degree of approximation
    ! optionNodes:  0 for equally spaced nodes
    !               1 for optimal nodes
    !
    ! NOTE: Ordering = vertices + sweep
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    REAL*8, INTENT(out)  :: nodes(2,*)
    INTEGER :: i, j, n
    REAL*8  :: h0

    IF( optionNodes==0)THEN
       n = 4
       h0 = 2.d0 / p

       ! Vertices
       nodes(:,1) = (/-1, -1/)
       nodes(:,2) = (/1, -1/)
       nodes(:,3) = (/-1, 1/)

       ! First edge
       DO i=1,p-1
          nodes(:,n) = (/-1+i*h0,  -1.0d0/)
          n = n + 1
       ENDDO

       ! Sweep
       DO j=1,p-1
          DO i=0,p-j
             nodes(:,n) = (/-1+i*h0,  -1+j*h0 /)
             n = n + 1
          ENDDO
       ENDDO

    ELSEIF (optionNodes==1)THEN
       CALL HighOrderCell_feketeNodes2D(p, nodes)
    ENDIF

  END SUBROUTINE HighOrderCell_SetNodesTRI


  FUNCTION HighOrderCell_trianglePermutation(p,iorder) RESULT(indx)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, iorder
    INTEGER, DIMENSION((p+1)*(p+2)/2) :: indx
    INTEGER, DIMENSION((p+1)*(p+2)/2) :: nodes, indp
    INTEGER :: n, i, j, k

    n = (p+1)*(p+2)/2

    IF(ABS(iorder)==1)THEN
       indx(1:n) = (/(i, i=1,n, 1)/)
       IF(iorder==1) RETURN
    ELSE IF(ABS(iorder)>=2)THEN
       !--- indx:  for (2,1,3)
       k = 0
       DO j = 1, p+1
          k = k+1
          indx(k) = p+2-j
          DO i = 2, p+2-j
             k = k+1
             indx(k) = indx(k-1) + p+2-i
          ENDDO
       ENDDO

       IF(ABS(iorder)==3)THEN
          nodes(1:n) = indx(1:n) 
          DO i = 1,n
             indx(i) = nodes(indx(i))
          ENDDO
       ENDIF
    ENDIF

    IF(iorder<0)THEN
       !    indp:  for (1,3,2)
       k = 0
       DO j = 1, p+1
          k = k+1
          indp(k) = j
          DO i = 2, p+2-j
             k = k+1
             indp(k) = indp(k-1) + p+3-i
          ENDDO
       ENDDO

       nodes(1:n) = indx(1:n) 
       DO i = 1,n
          indx(i) = nodes(indp(i))
       ENDDO
    ENDIF

    !--- put the vetices first

    DO i = 1,n
       IF(indx(i)==p+1)THEN
          indx(i) = 2
       ELSE IF(indx(i)==n)THEN
          indx(i) = 3
       ELSE IF(indx(i)>p+1)THEN
          indx(i) = indx(i)+1
       ELSE IF(indx(i)>1)THEN
          indx(i) = indx(i)+2
       ENDIF
    ENDDO

    j = 4
    k = 1
    DO i = 1, n
       IF(indx(i)>3)THEN
          nodes(j) = indx(i)
          j = j+1
       ELSE
          nodes(k) = indx(i)
          k = k+1
       ENDIF
    ENDDO

    indx(1:n) = nodes(1:n)

  END FUNCTION HighOrderCell_trianglePermutation


  SUBROUTINE HighOrderCell_SetNodesTET(p, optionNodes, Coord)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    REAL*8, INTENT(out)  :: Coord(3,*)
    INTEGER :: i, j, k, n
    REAL*8  :: h0

    IF( optionNodes==0)THEN

       h0 = 2.d0/(p)
       Coord(:,1) =   (/-1, -1, -1/)
       Coord(:,2) =   (/ 1, -1, -1/)
       Coord(:,3) =   (/ -1, 1, -1/)
       Coord(:,4) =   (/ -1, -1, 1/)
       n = 4
       DO i=1,p-1
          n = n+1
          Coord(:,n) = (/-1+i*h0,  -1.0d0, -1.0d0/)
       ENDDO

       DO j=1,p-1
          DO i=0,p-j
             n = n+1
             Coord(:,n) = (/-1+i*h0,  -1+j*h0, -1.0d0/)
          ENDDO
       ENDDO

       DO k=1,p-1
          DO j=0,p-k
             DO i=0,p-k-j
                n = n+1
                Coord(:,n) = (/-1+i*h0,  -1+j*h0, -1+k*h0/)
             ENDDO
          ENDDO
       ENDDO

    ELSE IF(optionNodes==1)THEN

       CALL HighOrderCell_feketeNodes3D(p, Coord)

    ENDIF

  END SUBROUTINE HighOrderCell_SetNodesTET


  SUBROUTINE HighOrderCell_getEdgeNodes_TRI(p, faceNodes)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    INTEGER, DIMENSION(p+1,3), INTENT(out)  :: faceNodes
    INTEGER :: nOfFaceNodes, nOfNodes, i

    nOfFaceNodes = p+1
    nOfNodes     = (p+1)*(p+2)/2

    ! Permutations (face nodes)
    faceNodes(1:2,1) = (/2, 3/)
    faceNodes(1:2,2) = (/3, 1/)
    faceNodes(1:2,3) = (/1, 2/)

    IF(p==1) RETURN

    faceNodes(3, 1:3) = (/3+2*(p-1)+1, nOfNodes-1, 4/)    
    DO i = 4, p+1
       faceNodes(i, 1) = faceNodes(i-1, 1) + p+3-i
       faceNodes(i, 2) = faceNodes(i-1, 2) + 1-i
       faceNodes(i, 3) = faceNodes(i-1, 3) + 1
    ENDDO

  END SUBROUTINE HighOrderCell_getEdgeNodes_TRI



  !
  ! faceNodes = HighOrderCell_computeFaceNodesTet(p)
  !
  ! Return a matrix 4 x nNodeFace containing ordered nodes for every face
  ! Nodal distribution is not defined by an input for more functionality
  ! Although equally-spaced nodal distribution is used the node numbering is
  ! the same for all nodal distributions
  SUBROUTINE  HighOrderCell_getFaceNodes_TET(p, faceNodes)
    ! Face nodes with permutations
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    INTEGER, DIMENSION((p+1)*(p+2)/2,4), INTENT(out)  :: faceNodes
    INTEGER :: iface, n, ND, pp, ip, id0, id1, id2
    REAL*8  :: TOL
    REAL*8, DIMENSION(3,(p+1)*(p+2)*(p+3)/6) :: Xi
    REAL*8, DIMENSION(4,(p+1)*(p+2)*(p+3)/6) :: xp
    REAL*8, DIMENSION((p+1)*(p+2)/2) :: Xf
    INTEGER, DIMENSION((p+1)*(p+2)/2) :: nodes, indx

    faceNodes(1:3, 1) = (/2,4,3/)
    faceNodes(1:3, 2) = (/3,4,1/)
    faceNodes(1:3, 3) = (/4,2,1/)
    faceNodes(1:3, 4) = (/1,2,3/)
    IF(p==1) RETURN

    ND = (p+1)*(p+2)*(p+3)/6
    CALL HighOrderCell_SetNodesTET(p, 0, Xi)

    ! Nodes of every face
    ! Face 1: xi+eta+zeta=-1
    ! Face 2: xi=-1
    ! Face 3: eta=-1
    ! Face 4: zeta=-1

    xp(1:3,:) = 0.5d0 * (1 + Xi(1:3,:))
    xp(4,  :) = 1.d0 - (Xp(1,:)+Xp(2,:)+Xp(3,:))

    TOL = 1.e-5
    pp = 2*p

    DO iface = 1,4
       DO id0 = 1,4
          IF(xp(id0, faceNodes(1,iface))>TOL ) CYCLE
          IF(xp(id0, faceNodes(2,iface))>TOL ) CYCLE
          IF(xp(id0, faceNodes(3,iface))>TOL ) CYCLE
          EXIT
       ENDDO
       DO id1 = 1,4
          IF(id1==id0) CYCLE
          IF(xp(id1, faceNodes(1,iface))>TOL ) CYCLE
          IF(xp(id1, faceNodes(3,iface))>TOL ) CYCLE
          EXIT
       ENDDO
       DO id2 = 1,4
          IF(id2==id0 .OR. id2==id1) CYCLE
          IF(xp(id2, faceNodes(1,iface))>TOL ) CYCLE
          IF(xp(id2, faceNodes(2,iface))>TOL ) CYCLE
          EXIT
       ENDDO

       n = 0
       DO ip = 5,ND
          IF(xp(id0, ip)>TOL ) CYCLE
          n = n+1
          nodes(n) = ip
          xf(n) = Xp(id1,ip) + Xp(id2,ip)*pp
       ENDDO
       IF(n/=(p+1)*(p+2)/2-3) STOP 'cnwoifh084er90'      

       CALL quicksort(n,xf,indx)
       DO ip = 1,n
          faceNodes(ip+3,iface) = nodes(indx(ip))
       ENDDO
    ENDDO

  END SUBROUTINE  HighOrderCell_getFaceNodes_TET


  !
  ! edgeNodes = HighOrderCell_computeFaceNodesTet(p)
  !
  ! Return a matrix 4 x nNodeFace containing ordered nodes for every face
  ! Nodal distribution is not defined by an input for more functionality
  ! Although equally-spaced nodal distribution is used the node numbering is
  ! the same for all nodal distributions
  SUBROUTINE  HighOrderCell_getEdgeNodes_TET(p, edgeNodes)
    ! Face nodes with permutations
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p
    INTEGER, DIMENSION(:,:), INTENT(out)  :: edgeNodes
    INTEGER :: iEdge, n, ND, ip, id0, id1, id2
    REAL*8  :: TOL
    REAL*8, DIMENSION(3,(p+1)*(p+2)*(p+3)/6) :: Xi
    REAL*8, DIMENSION(4,(p+1)*(p+2)*(p+3)/6) :: xp
    REAL*8, DIMENSION(p+1) :: Xf
    INTEGER, DIMENSION(p+1) :: nodes, indx

    edgeNodes(1:2, 1:6) = iEdge_Tet(1:2,1:6)
    IF(p==1) RETURN

    ND = (p+1)*(p+2)*(p+3)/6
    CALL HighOrderCell_SetNodesTET(p, 0, Xi)

    ! Nodes of every edge

    xp(1:3,:) = 0.5d0 * (1 + Xi(1:3,:))
    xp(4,  :) = 1.d0 - (Xp(1,:)+Xp(2,:)+Xp(3,:))

    TOL = 1.e-5

    DO iEdge = 1,6
       DO id0 = 1,4
          IF(xp(id0, edgeNodes(1,iEdge))>TOL ) CYCLE
          IF(xp(id0, edgeNodes(2,iEdge))>TOL ) CYCLE
          EXIT
       ENDDO
       DO id1 = 1,4
          if(id1==id0) cycle
          IF(xp(id1, edgeNodes(1,iEdge))>TOL ) CYCLE
          IF(xp(id1, edgeNodes(2,iEdge))>TOL ) CYCLE
          EXIT
       ENDDO
       DO id2 = 1,4
          IF(id2==id0) CYCLE
          IF(id2==id1) CYCLE
          IF(xp(id2, edgeNodes(1,iEdge))>TOL ) CYCLE
          EXIT
       ENDDO

       n = 0
       DO ip = 5,ND
          IF(xp(id0, ip)>TOL ) CYCLE
          IF(xp(id1, ip)>TOL ) CYCLE
          n = n+1
          nodes(n) = ip
          xf(n) = Xp(id2,ip)
       ENDDO
       IF(n/=p-1) STOP 'dfjweojrf23820'      

       CALL quicksort(n,xf,indx)
       DO ip = 1,n
          edgeNodes(ip+2,iEdge) = nodes(indx(ip))
       ENDDO
    ENDDO

  END SUBROUTINE  HighOrderCell_getEdgeNodes_TET



  FUNCTION HighOrderCell_get1DCell(p,optionNodes) RESULT(HighCell)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    TYPE(HighOrderCellType) :: HighCell

    HighCell%Nsd = 1
    HighCell%GridOrder = p
    HighCell%numCellNodes = (p+1)
    HighCell%numFaceNodes = 1
    HighCell%numVertices = 2
    HighCell%optionNodes = optionNodes

    ! Nodal distribution
    IF(optionNodes>=0)THEN
       ALLOCATE(HighCell%Coord(1,HighCell%numCellNodes))
       CALL HighOrderCell_SetNodes1D(p, optionNodes,HighCell%Coord(1,:))
    ENDIF

    ! Face nodes
    ALLOCATE(HighCell%faceNodes(1,2))
    HighCell%faceNodes(1,:) = (/2, 1/)

  END FUNCTION HighOrderCell_get1DCell



  FUNCTION HighOrderCell_getTRICell(p,optionNodes) RESULT(HighCell)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    TYPE(HighOrderCellType) :: HighCell

    HighCell%Nsd = 2
    HighCell%GridOrder = p
    HighCell%numCellNodes = (p+1)*(p+2)/2
    HighCell%numFaceNodes = p+1
    HighCell%numVertices = 3
    HighCell%optionNodes = optionNodes

    ! Nodal distribution
    IF(optionNodes>=0)THEN
       ALLOCATE(HighCell%Coord(2,HighCell%numCellNodes))
       CALL HighOrderCell_SetNodesTRI(p, optionNodes,HighCell%Coord)
    ENDIF

    ! Face nodes
    ALLOCATE(HighCell%faceNodes(HighCell%numFaceNodes,3))
    CALL HighOrderCell_getEdgeNodes_TRI(p, HighCell%faceNodes)

  END FUNCTION HighOrderCell_getTRICell


  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ! 3D
  !%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  FUNCTION HighOrderCell_getTETCell(p,optionNodes) RESULT( HighCell)
    IMPLICIT NONE
    INTEGER, INTENT(in) :: p, optionNodes
    TYPE(HighOrderCellType) :: HighCell

    HighCell%Nsd = 3
    HighCell%GridOrder = p
    HighCell%numCellNodes = (p+1)*(p+2)*(p+3)/6
    HighCell%numFaceNodes = (p+1)*(p+2)/2
    HighCell%numVertices = 4
    HighCell%optionNodes = optionNodes

    ! Nodal distribution
    IF(optionNodes>=0)THEN
       ALLOCATE(HighCell%Coord(3,HighCell%numCellNodes))
       CALL HighOrderCell_SetNodesTET(p, optionNodes, HighCell%Coord )
    ENDIF


    ! Face nodes
    ALLOCATE(HighCell%faceNodes(HighCell%numFaceNodes,4)) 
    CALL HighOrderCell_getFaceNodes_TET(p,  HighCell%faceNodes)

  END FUNCTION HighOrderCell_getTETCell


  !>
  !! Function to compute the shape functions (& derivatives) at Gauss points
  !!
  !! Input:
  !! GridOrder:  degree of interpolation
  !! coord: nodal coordinates at the reference element
  !!
  !! Output:
  !! Sh  : shape functions evaluated at the gauss points
  !! Dsh : shape function derivatives
  !! gaussWeights:   weights 
  !<
  SUBROUTINE HighOrderCell_calcGaussShape1D(HighCell)
    IMPLICIT NONE
    TYPE(HighOrderCellType), INTENT(INOUT) :: HighCell
    REAL*8, POINTER  ::  Sh(:), Dsh1(:)
    INTEGER :: NP, NG, allocateStatus, iG, NG1D
    REAL*8, ALLOCATABLE ::  x(:), w(:)

    NP   = HighCell%numCellNodes
    NG1D = HighCell%GridOrder + 1
    NG   = NG1D
    HighCell%numGaussNodes = NG

    ALLOCATE(x(NG1D), stat=allocateStatus)
    ALLOCATE(w(NG1D), stat=allocateStatus)
    CALL gaussLegendre1D(x, w, NG1D, -1.D0, 1.D0)

    ALLOCATE(Sh(NP), Dsh1(NP))
    ALLOCATE(HighCell%GaussShapeFunct(NP,  NG))
    ALLOCATE(HighCell%GaussShapeDeriv(NP,1,NG))
    ALLOCATE(HighCell%gaussWeights(NG))    
    HighCell%gaussWeights(:) = w(:)

    DO iG = 1,NG1D
       CALL orthoPolyDeriv1D(Sh,Dsh1, x(iG), HighCell%GridOrder)
       CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, sh)
       CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh1)
       HighCell%GaussShapeFunct(:,  iG) = Sh(:)
       HighCell%GaussShapeDeriv(:,1,iG) = DSh1(:)
    END DO

    DEALLOCATE(x, w, Sh, Dsh1,  stat=allocateStatus)
    IF(allocateStatus/=0) STOP "ERR->HighOrderCell_calcGaussShape1D: unable to deallocate"

  END SUBROUTINE HighOrderCell_calcGaussShape1D



  !>
  !! Function to compute the shape functions (& derivatives) at Gauss points
  !!
  !! Input:
  !! GridOrder:  degree of interpolation
  !! coord: nodal coordinates at the reference element
  !!
  !! Output:
  !! Sh  : shape functions evaluated at the gauss points
  !! Dsh : shape function derivatives (1:2) 
  !! gaussWeights:   weights 
  !<
  SUBROUTINE HighOrderCell_calcGaussShapeTri(HighCell)
    IMPLICIT NONE
    TYPE(HighOrderCellType), INTENT(INOUT) :: HighCell
    REAL*8, POINTER  ::  Sh(:), Dsh1(:), Dsh2(:)
    INTEGER :: NP, NG, allocateStatus, iGauss, iG, jG, NG1D
    REAL*8, ALLOCATABLE ::  x(:), w(:)
    REAL*8 :: xRST(2)

    NP   = HighCell%numCellNodes
    NG1D = HighCell%GridOrder + 1
    NG   = NG1D**2
    HighCell%numGaussNodes = NG

    ALLOCATE(x(NG1D), stat=allocateStatus)
    ALLOCATE(w(NG1D), stat=allocateStatus)
    CALL gaussLegendre1D(x, w, NG1D, -1.D0, 1.D0)

    ALLOCATE(Sh(NP), Dsh1(NP), Dsh2(NP))
    ALLOCATE(HighCell%GaussShapeFunct(NP,  NG))
    ALLOCATE(HighCell%GaussShapeDeriv(NP,2,NG))
    ALLOCATE(HighCell%gaussWeights(NG))    

    iGauss = 1
    DO iG = 1,NG1D
       DO jG = 1,NG1D
          xRST = (/ x(iG), x(jG) /)
          CALL orthoPolyDerivRST2D(Sh,Dsh1,Dsh2, xRST, HighCell%GridOrder)
          CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, sh)
          CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh1)
          CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh2)
          HighCell%GaussShapeFunct(:,  iGauss) = Sh(:)
          HighCell%GaussShapeDeriv(:,1,iGauss) = DSh1(:)
          HighCell%GaussShapeDeriv(:,2,iGauss) = DSh2(:)
          HighCell%gaussWeights(iGauss) =(w(iG)*w(jG))*((1-xRST(2))/2)
          iGauss = iGauss + 1
       END DO
    END DO

    DEALLOCATE(x, w, Sh, Dsh1, Dsh2, stat=allocateStatus)
    IF(allocateStatus/=0) STOP "ERR->HighOrderCell_calcGaussShapeTri: unable to deallocate"

  END SUBROUTINE HighOrderCell_calcGaussShapeTri


  !>
  !! Function to compute the shape functions (& derivatives) at Gauss points
  !!
  !! Input:
  !! GridOrder:  degree of interpolation
  !! coord: nodal coordinates at the reference element
  !!
  !! Output:
  !! Sh  : shape functions evaluated at the gauss points
  !! Dsh : shape function derivatives (1:3) 
  !! gaussWeights:   weights 
  !<
  SUBROUTINE HighOrderCell_calcGaussShapeTet(HighCell)
    IMPLICIT NONE
    TYPE(HighOrderCellType), INTENT(INOUT) :: HighCell
    REAL*8, POINTER  ::  Sh(:), Dsh1(:), Dsh2(:), Dsh3(:)
    INTEGER :: NP, NG, allocateStatus, iGauss, iG, jG, kG, NG1D
    REAL*8, ALLOCATABLE ::  x(:), w(:)
    REAL*8 :: xRST(3)

    NP   = HighCell%numCellNodes
    NG1D = HighCell%GridOrder + 1
    NG   = NG1D**3
    HighCell%numGaussNodes = NG

    ALLOCATE(x(NG1D), stat=allocateStatus)
    ALLOCATE(w(NG1D), stat=allocateStatus)
    CALL gaussLegendre1D(x, w, NG1D, -1.D0, 1.D0)

    ALLOCATE(Sh(NP), Dsh1(NP), Dsh2(NP), Dsh3(NP))
    ALLOCATE(HighCell%GaussShapeFunct(NP,  NG))
    ALLOCATE(HighCell%GaussShapeDeriv(NP,3,NG))
    ALLOCATE(HighCell%gaussWeights(NG))    

    iGauss = 1
    DO iG = 1,NG1D
       DO jG = 1,NG1D
          DO kG = 1,NG1D
             xRST = (/ x(iG), x(jG), x(kG) /)
             CALL orthoPolyDerivRST3D(Sh,Dsh1,Dsh2,Dsh3, xRST, HighCell%GridOrder)
             CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, sh)
             CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh1)
             CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh2)
             CALL Matrix_LU_Solver(NP, HighCell%shapeMax, HighCell%shapeLU, Dsh3)
             HighCell%GaussShapeFunct(:,  iGauss) = Sh(:)
             HighCell%GaussShapeDeriv(:,1,iGauss) = DSh1(:)
             HighCell%GaussShapeDeriv(:,2,iGauss) = DSh2(:)
             HighCell%GaussShapeDeriv(:,3,iGauss) = DSh3(:)
             HighCell%gaussWeights(iGauss) =(w(iG)*w(jG)*w(kG))*((1-xRST(2))/2)*((1-xRST(3))/2)**2
             iGauss = iGauss + 1
          END DO
       END DO
    END DO

    DEALLOCATE(x, w, Sh, Dsh1, Dsh2, Dsh3, stat=allocateStatus)
    IF(allocateStatus/=0) STOP "ERR->HighOrderCell_calcGaussShapeTet: unable to deallocate"

  END SUBROUTINE HighOrderCell_calcGaussShapeTet


  !>
  !!   Return the shape function at (u,v) for a cell.
  !!   @param[in]  uv  : -1<=uv(:)<=1
  !!   @param[out] Sh  : shape function at uv(:).
  !!
  !!   \f$   F(uv) = Sh(1)F(p_1)+Sh(2)F(p_2)+...+Sh(n)F(p_n)  \f$   \n
  !!     where \f$ (p_i) \f$ = are the position of cell nodes, and n = HighCell.numCellNodes.
  !!
  !<
  SUBROUTINE HighOrderCell_calcShapeFunctions(HighCell, uv, Sh )
    IMPLICIT NONE
    TYPE(HighOrderCellType), INTENT(IN) :: HighCell
    REAL*8,  INTENT(IN) :: uv(:)
    REAL*8,  INTENT(OUT) :: Sh(:)
    INTEGER :: NP, nsd

    NP  = HighCell%numCellNodes
    nsd = HighCell%Nsd
    IF(nsd>SIZE(uv,1))THEN
       WRITE(*,*)' Error--- nsd>SIZE(uv,1): ',nsd, SIZE(uv,1)
       STOP 'HighOrderCell_calcShapeFunctions::'
    ENDIF

    IF(nsd==2)THEN
       CALL orthoPoly2D(Sh,uv(1:2),HighCell%GridOrder)
    ELSE IF(nsd==3)THEN
       CALL orthoPoly3D(Sh,uv,HighCell%GridOrder)
    ENDIF

    CALL Matrix_LU_Solver(NP,HighCell%shapeMax,HighCell%shapeLU, Sh)

  END SUBROUTINE HighOrderCell_calcShapeFunctions


  !>
  !!   Return the derivative of shape function at uv(:) for cell.
  !!   @param[in]  uv  : -1<=uv(:)<=1
  !!   @param[out] Sh  : shape function at uv(:).
  !!   @param[out] Dsh : weights of derivative.
  !!
  !!   \f$  \frac{\partial}{\partial u} F(uv)
  !!         = Ds(1,1)F(p_1)+Ds(2,1)F(p_2)+...+Ds(n,1)F(p_n)  \f$   \n
  !!   \f$  \frac{\partial}{\partial v} F(uv)
  !!         = Ds(1,2)F(p_1)+Ds(2,2)F(p_2)+...+Ds(n,2)F(p_n)  \f$   \n
  !!     where \f$ (p_i) \f$ = are the position of cell nodes, and n = HighCell.numCellNodes.
  !!
  !<
  SUBROUTINE HighOrderCell_calcShapeFunctionsDeriv(HighCell, uv, Sh, Dsh)
    IMPLICIT NONE
    TYPE(HighOrderCellType), INTENT(IN) :: HighCell
    REAL*8,  INTENT(IN) :: uv(:)
    REAL*8,  INTENT(OUT) :: Sh(:), Dsh(:,:)
    INTEGER :: NP, nsd
    REAL*8 :: xRST(3)

    NP  = HighCell%numCellNodes
    nsd = HighCell%Nsd
    IF(nsd>SIZE(uv,1)) STOP 'sfk9wri293e321dew'

    IF(nsd==2)THEN
       CALL mapXiEtaZetaToRST2D(xRST(1:2), uv(1:2))
       CALL orthoPolyDerivRST2D(Sh,Dsh(:,1),Dsh(:,2),xRST,HighCell%GridOrder)
    ELSE IF(nsd==3)THEN
       CALL mapXiEtaZetaToRST3D(xRST, uv)
       CALL orthoPolyDerivRST3D(Sh,Dsh(:,1),Dsh(:,2),Dsh(:,3),xRST,HighCell%GridOrder)
    ENDIF

    CALL Matrix_LU_Solver(NP,HighCell%shapeMax,HighCell%shapeLU, Sh)
    CALL Matrix_LU_Solver(NP,HighCell%shapeMax,HighCell%shapeLU, Dsh(:,1))
    CALL Matrix_LU_Solver(NP,HighCell%shapeMax,HighCell%shapeLU, Dsh(:,2))
    IF(nsd==2)RETURN
    CALL Matrix_LU_Solver(NP,HighCell%shapeMax,HighCell%shapeLU, Dsh(:,3))

  END SUBROUTINE HighOrderCell_calcShapeFunctionsDeriv


  !>
  !! HighCell = createReferenceElement(p,nsd,optionNodes)
  !! creates the struct containing the reference element information
  !!
  !! Input:
  !! p:        degree of the interpolation
  !! nsd:      number of spatial dimensions
  !! optionNodes:   0 for equally spaced nodes
  !!                1 for optimal nodes
  !!               -1 for connextivity only, no coordinates.
  !<
  FUNCTION HighOrderCell_getHighCell(p,nsd,optionNodes) RESULT(HighCell )
    IMPLICIT NONE
    INTEGER, INTENT(IN) :: p, nsd, optionNodes
    TYPE(HighOrderCellType) :: HighCell
    REAL*8, DIMENSION(:,:), POINTER :: coordt
    INTEGER :: NP, NG, g

    IF( nsd==1 )THEN
       HighCell = HighOrderCell_get1DCell(p,optionNodes)
    ELSEIF( nsd == 2)THEN
       HighCell = HighOrderCell_getTRICell(p,optionNodes)
    ELSEIF( nsd == 3)THEN
       HighCell = HighOrderCell_getTETCell(p,optionNodes)
    ENDIF

    IF(optionNodes<0) RETURN

    !--- shape function help matrix

    NP  = HighCell%numCellNodes
    ALLOCATE(coordt(NP,nsd))
    coordt(1:NP, 1:nsd) = TRANSPOSE(HighCell%Coord(1:nsd, 1:NP))

    ALLOCATE(HighCell%shapeMax(NP,NP), HighCell%shapeLU(NP))
    IF(nsd==1)THEN
       CALL vandermondeMatrix1D(HighCell%shapeMax, HighCell%GridOrder, Coordt(:,1))
    ELSE IF(nsd==2)THEN
       CALL vandermondeMatrix2D(HighCell%shapeMax, HighCell%GridOrder, Coordt)
    ELSE IF(nsd==3)THEN
       CALL vandermondeMatrix3D(HighCell%shapeMax, HighCell%GridOrder, Coordt)
    ENDIF
    HighCell%shapeMax = TRANSPOSE(HighCell%shapeMax)
    CALL Matrix_LU_Decomposition(NP, HighCell%shapeMax, HighCell%shapeLU)

    DEALLOCATE(coordt)

    !--- shape function at gauss nodes

    IF(nsd==1)THEN
       CALL HighOrderCell_calcGaussShape1D(HighCell)
    ELSEIF(nsd==2)THEN
       CALL HighOrderCell_calcGaussShapeTri(HighCell)
    ELSEIF(nsd==3)THEN
       CALL HighOrderCell_calcGaussShapeTet(HighCell)
    ENDIF

  END FUNCTION HighOrderCell_getHighCell

  !>
  !!   Delete the cell strucutre and release memory.
  !<
  SUBROUTINE HighOrderCell_Clear(HighCell)
    TYPE(HighOrderCellType), INTENT(INOUT) :: HighCell
     HighCell%Nsd           = 0
     HighCell%GridOrder     = 1
     HighCell%numCellNodes  = 0
     HighCell%numFaceNodes  = 0
     HighCell%numVertices   = 0
     HighCell%optionNodes   = 1
     HighCell%numGaussNodes = 0
     IF(ASSOCIATED(HighCell%Coord))           DEALLOCATE(HighCell%Coord)
     IF(ASSOCIATED(HighCell%faceNodes))       DEALLOCATE(HighCell%faceNodes)
     IF(ASSOCIATED(HighCell%gaussShapeFunct)) DEALLOCATE(HighCell%gaussShapeFunct)
     IF(ASSOCIATED(HighCell%gaussShapeDeriv)) DEALLOCATE(HighCell%gaussShapeDeriv)
     IF(ASSOCIATED(HighCell%gaussWeights))    DEALLOCATE(HighCell%gaussWeights)
     IF(ASSOCIATED(HighCell%shapeMax))        DEALLOCATE(HighCell%shapeMax)
     IF(ASSOCIATED(HighCell%shapeLU))         DEALLOCATE(HighCell%shapeLU)     
  END SUBROUTINE HighOrderCell_Clear
  
  
  
END MODULE HighOrderCell
